/*
ATLSWC - AddTLSSection With Code
Copyright (C) 2012  George Nicolaou nicolaou.george{at}gmail{dot}com

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <Windows.h>
#include <stdlib.h>
#pragma optimize("", off)
#include "include.h"
#include "pe_stuff.h"
#include "memory.h"

void * GetEntryPointAddress( PFILE_IN_MEMORY pFile )
{
	PIMAGE_NT_HEADERS pNtHeaders;
	if( ( pNtHeaders = GetNtHeaders( pFile ) ) != NULL )
		return (void *)pNtHeaders->OptionalHeader.AddressOfEntryPoint;
	else
		return NULL;
}

void * VirtualAddressToFilePtrReal( 
	void * lpAddress, 
	PFILE_IN_MEMORY pFile )
{
	PIMAGE_NT_HEADERS pNtHeaders;
	PIMAGE_SECTION_HEADER pSection;
	unsigned char * lpucAddrOffset;
	unsigned long ulSectionEndAddr;

	pNtHeaders = GetNtHeaders( pFile );
	pSection = (PIMAGE_SECTION_HEADER)( (char *)pNtHeaders + 
		sizeof( IMAGE_NT_HEADERS ) );
	while( pSection ) {
		if( pSection->VirtualAddress <= (DWORD)lpAddress ) {
			ulSectionEndAddr = pSection->SizeOfRawData + 
				pSection->VirtualAddress;
			if( ulSectionEndAddr >= (DWORD)lpAddress ) {
				lpucAddrOffset = (unsigned char *)(
					(char *)lpAddress -
					pSection->VirtualAddress +
					pSection->PointerToRawData );
				return (void *)
					((char *)pFile->lpFileBuffer + (int)lpucAddrOffset );
			}
		}
		pSection++;
	}
		return NULL;
}

void * AddTLSToPE( 
	PFILE_IN_MEMORY pFile, 
	int iSizeRequested, 
	unsigned char * lpucCode, 
	int iSizeOfCode )
{
	PIMAGE_SECTION_HEADER pSection;
	PIMAGE_TLS_DIRECTORY psTLSDirectory;
	PIMAGE_DATA_DIRECTORY pDataDirTLSEntry;
	PIMAGE_NT_HEADERS pNtHeaders;
	DWORD * pdwArray;
	DWORD * dwRealAddress;
	unsigned char * lpucCodeOffset;
	unsigned char * lpucRealCodeOffset;
	int i;

	if( iSizeRequested < iSizeOfCode ) {
		error_print( "Size Requested is Greater than size of code", NULL );
	}

	pSection = AddNewPESection( 
		pFile, 
		".tls", 
		iSizeRequested,
		//IMAGE_SCN_MEM_EXECUTE | 
		IMAGE_SCN_CNT_CODE | 
		IMAGE_SCN_MEM_READ | 
		IMAGE_SCN_MEM_WRITE |
		IMAGE_SCN_CNT_INITIALIZED_DATA
		);

	if( pSection == NULL ) {
		return NULL;
	}

	pNtHeaders = GetNtHeaders( pFile );

	psTLSDirectory = (PIMAGE_TLS_DIRECTORY) ( 
		(char *)pFile->lpFileBuffer + pSection->PointerToRawData );
	pdwArray = (DWORD *)(psTLSDirectory + 1);

	for( i = 0; i < 5; i++ ) {
		pdwArray[i] = (DWORD)NULL;
	}

	dwRealAddress = (DWORD *)(
		pSection->VirtualAddress + 
		pNtHeaders->OptionalHeader.ImageBase +
		sizeof(IMAGE_TLS_DIRECTORY));

	

	psTLSDirectory->StartAddressOfRawData = (DWORD)dwRealAddress++; //&pdwArray[0];
	psTLSDirectory->EndAddressOfRawData = (DWORD)dwRealAddress++; //&pdwArray[1];
	psTLSDirectory->AddressOfIndex = (DWORD)dwRealAddress++; //&pdwArray[2];
	psTLSDirectory->AddressOfCallBacks = (DWORD)dwRealAddress++; //&pdwArray[3];
	psTLSDirectory->SizeOfZeroFill = (DWORD)NULL;
	psTLSDirectory->Characteristics = (DWORD)NULL;

	lpucRealCodeOffset = (unsigned char *)(dwRealAddress + 1);
	lpucCodeOffset = (unsigned char *)&pdwArray[5];
	pdwArray[3] = (DWORD)lpucRealCodeOffset;
	//pdwArray[3] = 0x00401000;
	memcpy( lpucCodeOffset, lpucCode, iSizeOfCode );

	
	pDataDirTLSEntry = (PIMAGE_DATA_DIRECTORY)
		&pNtHeaders->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_TLS];

	pDataDirTLSEntry->VirtualAddress = (DWORD)pSection->VirtualAddress;
	pDataDirTLSEntry->Size = sizeof( IMAGE_TLS_DIRECTORY );
	
	pSection = (PIMAGE_SECTION_HEADER)((unsigned long)pNtHeaders + 
		sizeof( IMAGE_NT_HEADERS ));

	for( i = 0; i < pNtHeaders->FileHeader.NumberOfSections; i++ ) {
		if( pSection[i].Characteristics & IMAGE_SCN_CNT_CODE || 
			pSection[i].Characteristics & IMAGE_SCN_MEM_EXECUTE ) {
				pSection[i].Characteristics |= IMAGE_SCN_MEM_WRITE;
		}
	}

	return lpucCodeOffset;
}

PIMAGE_NT_HEADERS GetNtHeaders( PFILE_IN_MEMORY pFile )
{
	PIMAGE_DOS_HEADER pDosHeader;
	PIMAGE_NT_HEADERS pNtHeader;
	pDosHeader = (PIMAGE_DOS_HEADER)pFile->lpFileBuffer;
	if( pDosHeader->e_magic != IMAGE_DOS_SIGNATURE ) {
		error_print( "Invalid DOS Signature", NULL );
	}

	pNtHeader = (PIMAGE_NT_HEADERS)( 
		(char *)pFile->lpFileBuffer + 
		pDosHeader->e_lfanew );

	if( pNtHeader->Signature != IMAGE_NT_SIGNATURE ) {
		error_print( "Invalid PE Signature", NULL );
	}
	return pNtHeader;
}

PIMAGE_SECTION_HEADER AddNewPESection( 
	PFILE_IN_MEMORY lpFile, 
	char * lpszSectionName, 
	int nSectionSize,
	DWORD dwCharacteristics )
{
	PIMAGE_NT_HEADERS pNtHeader;
	PIMAGE_SECTION_HEADER pPrevSectionHeader;
	PIMAGE_SECTION_HEADER pNewSectionHeader;
	DWORD dwSizeOfRawData;
	DWORD dwVirtualAddress;
	DWORD dwPointerToRawData;
	DWORD dwRealVirtualSize;
	void * lpFileBuffer;

	lpFileBuffer = lpFile->lpFileBuffer;

	if( ( pNtHeader = GetNtHeaders( lpFile ) ) == NULL )
		return NULL;

	pPrevSectionHeader = (PIMAGE_SECTION_HEADER)( 
		(char *)pNtHeader + 
		sizeof(IMAGE_NT_HEADERS) + 
		sizeof(IMAGE_SECTION_HEADER) *
		( pNtHeader->FileHeader.NumberOfSections - 1 ) );

	pNewSectionHeader = pPrevSectionHeader + 1;

	pNtHeader->FileHeader.NumberOfSections++;

	//Size needed to write the actual TLS Directory
	nSectionSize = nSectionSize + sizeof( IMAGE_TLS_DIRECTORY ) + 
		sizeof(DWORD) * 5;

	dwSizeOfRawData = 
		nSectionSize + 
		pNtHeader->OptionalHeader.FileAlignment - 
		( nSectionSize % pNtHeader->OptionalHeader.FileAlignment);
	
	dwVirtualAddress = 
		pPrevSectionHeader->VirtualAddress +
		pPrevSectionHeader->SizeOfRawData + 
		pNtHeader->OptionalHeader.SectionAlignment - 
		( dwSizeOfRawData % pNtHeader->OptionalHeader.SectionAlignment );

	dwPointerToRawData = 
		pPrevSectionHeader->PointerToRawData + 
		pPrevSectionHeader->SizeOfRawData;

	dwRealVirtualSize = nSectionSize +
		pNtHeader->OptionalHeader.SectionAlignment -
		( nSectionSize % pNtHeader->OptionalHeader.SectionAlignment );

	if( strlen( lpszSectionName ) >= 8) {
		lpszSectionName[8] = 0;
		printf( 
			"Section Name more than 8 bytes.\nTrimming down to: %s",
			lpszSectionName );
	}

	memcpy( &pNewSectionHeader->Name, lpszSectionName, 8 );
	pNewSectionHeader->VirtualAddress = dwVirtualAddress;
	pNewSectionHeader->Misc.VirtualSize = nSectionSize;
	pNewSectionHeader->PointerToRawData = dwPointerToRawData;
	pNewSectionHeader->SizeOfRawData = dwSizeOfRawData;
	pNewSectionHeader->Characteristics = dwCharacteristics;
	//XXX Relocation pointers

	lpFile->lpFileBuffer = realloc( lpFile->lpFileBuffer, lpFile->lSize + 
		dwSizeOfRawData );
	lpFile->lSize += dwSizeOfRawData;

	pNtHeader = GetNtHeaders( lpFile );
	pNewSectionHeader = (PIMAGE_SECTION_HEADER)( 
		(char *)pNtHeader + 
		sizeof(IMAGE_NT_HEADERS) + 
		sizeof(IMAGE_SECTION_HEADER) *
		( pNtHeader->FileHeader.NumberOfSections - 1 ) );

	
	pNtHeader->OptionalHeader.SizeOfImage += dwRealVirtualSize;
	memset( ( (char *)lpFile->lpFileBuffer + 
		pNewSectionHeader->PointerToRawData ), 0, dwSizeOfRawData );

	return pNewSectionHeader;
}